/// @file   Application.cpp
/// @author Armin Co
///

#include <GLFW/glfw3.h>

#include "Application.hpp"
#include "Log.hpp"
#include "ApplicationEvent.hpp"


namespace GuiCore
{
    Application* Application::s_instance = nullptr;

    Application::Application(const std::string &name, uint32_t width, uint32_t height)
    {
        if (s_instance == nullptr)
        {
            Log::setup();
        }
        s_instance = this;
        m_window = std::unique_ptr<IWindow>(IWindow::create({name, width, height}));
        m_window->setEventCallback([this](auto &&... args) -> decltype(auto) { return this->onEvent(std::forward<decltype(args)>(args)...); });

        m_guiLayer = std::make_shared<ImGuiLayer>();
        pushLayer(m_guiLayer);
    }

    void Application::pushLayer(std::shared_ptr<Layer> layer)
    {
        m_layerStack.pushLayer(layer);
    }

    void Application::popLayer(std::shared_ptr<Layer> layer)
    {
        m_layerStack.popLayer(layer);
    }

    void Application::onEvent(Event &event)
    {
        Log::get()->trace("{}", event.toString());
        EventDispatcher dispatcher(event);
        dispatcher.dispatch<WindowCloseEvent>(
            [this](auto &&... args) -> decltype(auto) { return this->onWindowClose(std::forward<decltype(args)>(args)...); });
        dispatcher.dispatch<WindowResizeEvent>(
            [this](auto &&... args) -> decltype(auto) { return this->onWindowResize(std::forward<decltype(args)>(args)...); });
        
        for (auto it = m_layerStack.rbegin(); it != m_layerStack.rend(); ++it)
        {
            if (event.handled)
            {
                break;
            }
            (*it)->onEvent(event);
        }
    }

    void Application::run()
    {
        while (m_isRunnig)
        {
            float time = static_cast<float>(glfwGetTime());
            Timestep timeStep = time - m_lastFrameTime;
            m_lastFrameTime = time;

            if (!m_minimzed)
            {
                for (auto layer : m_layerStack)
                {
                    layer->onUpdate(timeStep);
                }
            }

            m_guiLayer->begin();
            for (auto layer: m_layerStack)
            {
                layer->onGuiRender();
            }
            m_guiLayer->end();

            m_window->onUpdate();
        }
    }

    bool Application::onWindowClose(Event &event)
    {
        m_isRunnig = false;
        return true;
    }

    bool Application::onWindowResize(Event &event)
    {
        WindowResizeEvent *rsEvent = static_cast<WindowResizeEvent*>(&event);
        Log::get()->trace("Window minimized, {0}, {1}", rsEvent->getWidth(), rsEvent->getHeight());
        glViewport(0,0, rsEvent->getWidth(), rsEvent->getHeight());

        return false;
    }

    Application& Application::get()
    {
        return *s_instance;
    }

    IWindow& Application::getWindow()
    {
        return *m_window;
    }
}

