/// @file   Shader.cpp
/// @author Armin Co
///

#include <fstream>
#include <sstream>

#include "Shader.hpp"
#include "Log.hpp"



namespace GuiCore
{
    static std::string readFileAsString(const std::string &filepath)
    {
        std::string result;
        std::ifstream in(filepath, std::ios::in);
        if (in.is_open())
        {
            std::stringstream ss;
            ss << in.rdbuf();
            result = ss.str();
        }
        else
        {
            Log::get()->trace("Shader::readFileAsString");
            Log::get()->error("Could not open file '{0}'", filepath);
        }
        return result;
    }

    Shader::~Shader()
    {
        glDeleteProgram(m_rendererID);
    }

    GLuint Shader::compileShader(GLenum type, const std::string &source)
    {
        GLuint shader = glCreateShader(type);

        const GLchar *sourceCStr = source.c_str();
        Log::get()->trace("Shader::compileShader");
        // Log::get()->debug("{0}", sourceCStr);
        glShaderSource(shader, 1, &sourceCStr, nullptr);

        glCompileShader(shader);

        GLint isCompiled = 0;
        glGetShaderiv(shader, GL_COMPILE_STATUS, &isCompiled);
        if (isCompiled == GL_FALSE)
        {
            GLint maxLength = 0;
            glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &maxLength);

            std::vector<GLchar> infoLog(maxLength);
            glGetShaderInfoLog(shader, maxLength, &maxLength, &infoLog[0]);

            glDeleteShader(shader);

            Log::get()->trace("Shader::compileShader");
            Log::get()->error("{0}", infoLog.data());
        }
        return shader;
    }

    Shader* Shader::fromGLSLTextFiles(const std::string& vertexShaderPath, const std::string &fragmentShaderPath)
    {
        Shader *shader = new Shader();
        shader->loadFromGLSLTextFiles(vertexShaderPath, fragmentShaderPath);
        return shader;
    }

    void Shader::loadFromGLSLTextFiles(const std::string &vertexShaderPath, const std::string &fragmentShaderPath)
    {
        std::string vertexSource = readFileAsString(vertexShaderPath);
        std::string fragmentSource = readFileAsString(fragmentShaderPath);

        GLuint program = glCreateProgram();

        GLuint vertexShader = compileShader(GL_VERTEX_SHADER, vertexSource);
        glAttachShader(program, vertexShader);
        GLuint fragmentShader = compileShader(GL_FRAGMENT_SHADER, fragmentSource);
        glAttachShader(program, fragmentShader);

        glLinkProgram(program);

        GLint isLinked = 0;
        glGetProgramiv(program, GL_LINK_STATUS, (int*)&isLinked);
		if (isLinked == GL_FALSE)
		{
			GLint maxLength = 0;
			glGetProgramiv(program, GL_INFO_LOG_LENGTH, &maxLength);

			std::vector<GLchar> infoLog(maxLength);
			glGetProgramInfoLog(program, maxLength, &maxLength, &infoLog[0]);

			glDeleteProgram(program);

			glDeleteShader(vertexShader);
			glDeleteShader(fragmentShader);

            Log::get()->trace("Shader::loadFromGLSLTextFiles");
			Log::get()->error("{0}", infoLog.data());
		}

        glDetachShader(program, vertexShader);
		glDetachShader(program, fragmentShader);
		glDeleteShader(vertexShader);
		glDeleteShader(fragmentShader);

		m_rendererID = program;
    }

    GLuint Shader::getRendererID()
    {
        return m_rendererID;
    }
}
