/// @file   GameLayer.cpp
/// @author Armin Co
///

#include <GL/gl3w.h>
#include <GLFW/glfw3.h>

#include <imgui/imgui.h>

#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include <GuiCore/Log.hpp>
#include <GuiCore/Input.hpp>
#include <GuiCore/Glyphs.hpp>

#include "GameLayer.hpp"
#include "Log.hpp"

#include <string>
#include <thread>



double QPong::GameLayer::GameLayer::Bat::s_speedStepSize = 0.005;

template <typename T>
const T pow2(const T v)
{
    return v * v;
}

QPong::GameLayer::GameLayer::Bat::Bat(double x, double y, int keyUp, int keyDown)
    : m_x{x}
    , m_y{y}
    , m_vx{0.0}
    , m_vy{0.0}
    , m_height{0.21}
    , m_width{0.0}
    , m_keyUp{keyUp}
    , m_keyDown{keyDown}
{

}

void QPong::GameLayer::GameLayer::Bat::onUpdate(GuiCore::Timestep ts)
{
    constexpr double speed = 20.0;
    if (GuiCore::Input::isKeyPressed(m_keyDown))
    {
        m_vy += s_speedStepSize * speed; 
    }
    else if (GuiCore::Input::isKeyPressed(m_keyUp))
    {
        m_vy -= s_speedStepSize * speed;
    }
    else
    {
        m_vy = 0.0;
    }

    constexpr double maxSpeedFactor = 300.0;
    if (m_vy > s_speedStepSize * maxSpeedFactor)
    {
        m_vy = s_speedStepSize * maxSpeedFactor;
    }
    if (m_vy < -s_speedStepSize * maxSpeedFactor)
    {
        m_vy = -s_speedStepSize * maxSpeedFactor;
    }
    m_y += (m_vy * ts.getSeconds());

    constexpr double maxY {1.3};
    if (m_y > maxY)
    {
        m_y = maxY;
    }
    if (m_y < -maxY)
    {
        m_y = -maxY;
    }
}

double QPong::GameLayer::Bat::getPotential(double x, double y) const
{
    y -= m_y;
    double dist = 0.0;
    double scale = 0.0000001;

    if (y >= m_height)
    {
        dist = pow2(x - m_x) + pow2(y - m_height) + m_width;
    }
    else if (y <= (-m_height))
    {
        dist = pow2(x - m_x) + pow2((-y) - m_height) + m_width;
    }
    else
    {
        dist = pow2(m_x - x) + m_width;
    }
    return 1.0 / (pow2(dist) ) * scale;
}

double QPong::GameLayer::Bat::getX() const
{
    return m_x;
}

QPong::GameLayer::GameLayer(QPong::GameLayer::Properties props)
    :Layer("GameLayer")
    , m_vertexShaderPath{"assets/shaders/test.vert.glsl"}
    , m_fragmentShaderPath{"assets/shaders/test.frag.glsl"}
    , m_arraySize{props.resolution}
    , m_propagate{false}
    , m_leftBat(-0.8, 0.0, Key::W, Key::S)
    , m_rightBat(0.8, 0.0, Key::Up, Key::Down)
{
    m_arrayElements = m_arraySize * m_arraySize;
    m_numberOfQuads = (m_arraySize - 1) * (m_arraySize - 1);
    m_numberOfIndices = m_numberOfQuads * 6;

    m_initialParticleProperties.position.x = 0.0;
    m_initialParticleProperties.position.y = 0.0;
    m_initialParticleProperties.position.z = 0.0;
    m_initialParticleProperties.momentum.x = 1.0;
    m_initialParticleProperties.momentum.y = 0.5;
    m_initialParticleProperties.momentum.z = 0.00;
    m_initialParticleProperties.pointUnsharpness = 60.0;
    m_initialParticleProperties.startValue = 0.01;
    m_initialParticleProperties.hbar = 0.0002;

}

void QPong::GameLayer::calculatePointPositions(float *positions)
{
    float xMin {-3.0f};
    float xMax {0.0f};
    float yMin {-1.5f};
    float yMax {1.5f};
    float dx = (xMax - xMin) / (m_arraySize - 1.0f); // "spacing"
    float dy = (yMax - yMin) / (m_arraySize - 1.0f);
    {
        uint32_t i {0}; // array index
        for (int yId{0}; yId < m_arraySize; ++yId)
        {
            for (int xId{0}; xId < m_arraySize; ++xId)
            {
                positions[i++] = xMin + (static_cast<float>(xId) * dx); // x-Pos
                positions[i++] = yMax - (static_cast<float>(yId) * dy); // y-Pos
                positions[i++] = 0.0f;             // z-Pos
            }
        }
    }
}

void QPong::GameLayer::calculateMomentumViewPositions(float *positions)
{
    float xMin {0.2f};
    float xMax {2.2f};
    float yMin {1.0f};
    float yMax {-1.0f};
    float dx = (xMax - xMin) / (m_arraySize - 1.0f); // "spacing"
    float dy = (yMax - yMin) / (m_arraySize - 1.0f);
    {
        uint32_t i {0}; // array index
        for (int yId{0}; yId < m_arraySize; ++yId)
        {
            for (int xId{0}; xId < m_arraySize; ++xId)
            {
                positions[i++] = xMin + (static_cast<float>(xId) * dx); // x-Pos
                positions[i++] = yMax - (static_cast<float>(yId) * dy); // y-Pos
                positions[i++] = 0.0f;             // z-Pos
            }
        }
    }
}

void QPong::GameLayer::calculateTriangleIndices(uint32_t *indices)
{
    uint32_t i{0};
    for (int row {0}; row < (m_arraySize - 1); ++row)
    {
        for (int k {0}; k < (m_arraySize - 1); ++k)
        {
            int offset = row * m_arraySize;
            indices[i++] = offset + k;
            indices[i++] = offset + k + 1;
            indices[i++] = offset + k + m_arraySize;
            indices[i++] = offset + k + 1;
            indices[i++] = offset + k + m_arraySize;
            indices[i++] = offset + k + m_arraySize + 1;
        }
    }
}

void QPong::GameLayer::calculateBorderPotential(float *potential)
{
    double y {0.0};
    for (int i {0}; i < m_arrayElements; ++i)
    {
        y = yAtIndex(i);
        float pot = pow2(y) * pow2(y) * pow2(y) * pow2(y) * pow2(y) * pow2(y) * pow2(y) * 0.02;
        potential[i] = pot;
    }
}

double QPong::GameLayer::posAt(int pos) const
{
    return 2.0 * pos / m_arraySize - 1.0;
}

double  QPong::GameLayer::xAtIndex(int i) const
{
    return posAt(i % m_arraySize);
}

double QPong::GameLayer::yAtIndex(int i) const
{
    return posAt(i / m_arraySize);
}

     
double QPong::GameLayer::potentialAt(int pos) const
{
    if (pos > m_arraySize / 2)
    {
        pos -= m_arraySize;
    }
    return pos * M_PI * m_initialParticleProperties.hbar;
}

double QPong::GameLayer::pxAtIndex(int i) const
{
    return potentialAt(i % m_arraySize);
}

double QPong::GameLayer::pyAtIndex(int i) const
{
    return potentialAt(i / m_arraySize);
}



void QPong::GameLayer::initialiseParticleMomentum()
{
    constexpr std::complex<double> ci {0.0, 1.0}; // complex number
    const double hbar = m_initialParticleProperties.hbar;
    const double A0 = m_initialParticleProperties.startValue;
    const double px0 = m_initialParticleProperties.momentum.x / 20.0;
    const double py0 = m_initialParticleProperties.momentum.y / 20.0;
    const double x0 = m_initialParticleProperties.position.x;
    const double y0 = m_initialParticleProperties.position.y;
    const double lambda = m_initialParticleProperties.pointUnsharpness;

    double sum {0.0};
    m_particleAbsouluteSumStart = 0.0;
    double y{0.0}; 
    double x{0.0}; 
    for (auto i {0}; i < m_arrayElements; ++i)
    {
        x = xAtIndex(i);
        y = yAtIndex(i);
        m_psi[i] = A0 * exp( ci / hbar * ( x * px0 + y * py0) - lambda * (pow2(x-x0) + pow2(y-y0)));
        sum += pow2(std::real(m_psi[i])) + pow2(std::imag(m_psi[i]));
    }
    for (auto i {0}; i < m_arrayElements; ++i)
    {
        m_psi[i] *= (10.0 * (m_arraySize / 512.0)) / sqrt(sum);
        m_particleAbsouluteSumStart += pow2(std::real(m_psi[i])) + pow2(std::imag(m_psi[i]));
    }
    Log::get()->trace("Particle initialised");
}

double QPong::GameLayer::absorbParticle(int i, double c)
{
    double normPrevious = pow2(std::real(m_psi[i])) + pow2(std::imag(m_psi[i]));
    double absorb = pow2(std::cos(c * 10));
    m_psi[i] *= absorb;
    double delta = normPrevious - (pow2(std::real(m_psi[i])) + pow2(std::imag(m_psi[i])));
    return delta;
}

void QPong::GameLayer::applyPotentials(int indexBegin, int indexEnd, double dt)
{
    constexpr std::complex<double> ci {0.0, 1.0};
    const double hbar = m_initialParticleProperties.hbar;
    double x {0.0};
    double y {0.0};
    const double potential = 1.0;
    for (int i {indexBegin}; i < indexEnd; ++i)
    {
        double potSum = 1.0;
        if (m_potentialsEnabled)
        {
            potSum = 0.0;
            x = xAtIndex(i);
            y = yAtIndex(i);
            if (m_removeParticleEnabled)
            {    
                if (x < m_leftBat.getX())
                {
                    double distanceToEdge = -m_leftBat.getX() + x;
                    double c = distanceToEdge * (M_PI / 2.0) / (1.0 + m_leftBat.getX());
                    m_playerTwo.currentlyScored += absorbParticle(i, c);
                }
                if (x>m_rightBat.getX())
                {
                    double distanceToEdge = x - m_rightBat.getX();
                    double c = distanceToEdge * (M_PI / 2.0) / (1.0 + m_rightBat.getX());
                    m_playerOne.currentlyScored += absorbParticle(i, c);
                }
            }
            double potLeftBat = m_leftBat.getPotential(x, y);
            double potRightBat = m_rightBat.getPotential(x, y);
            potSum = potLeftBat + potRightBat + m_staticBorderPotential[i];
        }
        auto psi = m_psi[i] * exp(-ci * dt/hbar * potential * potSum);
        m_psi[i] = psi;
        if (m_potentialsEnabled)
        {
            m_dynamicPotential[i] = potSum;
        }
    }
}

void QPong::GameLayer::moveForward(int indexBegin, int indexEnd, double dt)
{
    constexpr std::complex<double> ci {0.0, 1.0};
    const double hbar = m_initialParticleProperties.hbar;
    double px {0.0};
    double py {0.0};
    constexpr double m = 1.0;   
    const double N = 1.0 / static_cast<double>(m_arrayElements);
    for (int i{indexBegin}; i<indexEnd; ++i)
    {
        px = pxAtIndex(i);
        py = pyAtIndex(i);
        double E = (pow2(px) + pow2(py)) / (2.0 * m);
        m_psi[i] *= exp(-ci * dt / hbar * E) * N;
    }
}

void QPong::GameLayer::updateMomentumView(int indexBegin, int indexEnd)
{
    int x {0};
    int y {0};
    for (int i {indexBegin}; i< indexEnd; ++i)
    {
        x = i % m_arraySize;
        y = i / m_arraySize;
        auto halfArraySize = m_arraySize / 2;
        
        if (y > halfArraySize)
        {
            y = y - halfArraySize;
        }
        else
        {
            y = y + halfArraySize;
        }

        if (x > halfArraySize)
        {
            x = x - halfArraySize;
        }
        else
        {
            x = x + halfArraySize;
        }
        m_momentum[i] = m_psi[(y * m_arraySize )- x] * 100.0;
    }
}

void QPong::GameLayer::propagateStep(double dt)
{
    std::vector<std::thread> workers;
    workers.reserve(m_renderingThreadsNumber);
    for (auto workerId{0}; workerId < m_renderingThreadsNumber; ++workerId)
    {
        int begin = workerId * (m_arrayElements / m_renderingThreadsNumber);
        int end = (workerId + 1) * (m_arrayElements / m_renderingThreadsNumber);
        workers.emplace_back(&QPong::GameLayer::applyPotentials, this, begin, end, dt);
    }
    for (auto &worker : workers)
    {
        worker.join();
    }
    fftw_execute(m_planBackwards);
    workers.clear();
    for (auto workerId{0}; workerId < m_renderingThreadsNumber; ++workerId)
    {
        int begin = workerId * (m_arrayElements / m_renderingThreadsNumber);
        int end = (workerId + 1) * (m_arrayElements / m_renderingThreadsNumber);
        workers.emplace_back(&QPong::GameLayer::moveForward, this, begin, end, dt);
    }
    for (auto &worker : workers)
    {
        worker.join();
    }
    if (m_renderMomentum)
    {   
        workers.clear();
        for (auto workerId{0}; workerId < m_renderingThreadsNumber; ++workerId)
        {
            int begin = workerId * (m_arrayElements / m_renderingThreadsNumber);
            int end = (workerId + 1) * (m_arrayElements / m_renderingThreadsNumber);
            workers.emplace_back(&QPong::GameLayer::updateMomentumView, this, begin, end);
        }
        for (auto &worker : workers)
        {
            worker.join();
        }
    }

    fftw_execute(m_planForward);
}

void QPong::GameLayer::onAttach()
{
    GuiCore::Glyphs::setup();

    m_renderMomentum = false;
    m_removeParticleEnabled = true;
    m_potentialsEnabled = true;
    m_playerOne.totalScore = 0;
    m_playerTwo.totalScore = 0;

    m_shader = GuiCore::Shader::fromGLSLTextFiles(
        m_vertexShaderPath,
        m_fragmentShaderPath
    );

    // static verticie positions
    m_particleViewPositions = new float[m_arrayElements * 3];
    calculatePointPositions(m_particleViewPositions);

    // Vertex array to store all buffers
    glGenVertexArrays(1, &m_vertexArray);
    glBindVertexArray(m_vertexArray);

    // Buffer of all point locations
    glGenBuffers(1, &m_pointsBuffer);
    glBindBuffer(GL_ARRAY_BUFFER, m_pointsBuffer);
    glBufferData(GL_ARRAY_BUFFER, sizeof(float) * 3 * (m_arrayElements), m_particleViewPositions, GL_DYNAMIC_DRAW);
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(float) * 3, 0);

    // Setup indies buffer
    uint32_t *indices = new uint32_t[m_numberOfIndices];
    calculateTriangleIndices(indices);

    glGenBuffers(1, &m_indicesBuffer);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_indicesBuffer);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(uint32_t) * m_numberOfIndices, indices, GL_STATIC_DRAW);
    delete(indices);

    // Setup particle buffer
    m_memorySize = sizeof(fftw_complex) * m_arrayElements;
    m_psi = static_cast<std::complex<double>*>(fftw_malloc(m_memorySize));
    m_momentum = static_cast<std::complex<double>*>(fftw_malloc(m_memorySize));

    glGenBuffers(1, &m_colorBuffer);
    glBindBuffer(GL_ARRAY_BUFFER, m_colorBuffer);
    glBufferData(GL_ARRAY_BUFFER, m_memorySize, nullptr, GL_DYNAMIC_DRAW);
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(1, 2, GL_DOUBLE, GL_FALSE, sizeof(fftw_complex), 0);

    // Setup momentum view
    m_momentumViewPositions = new float[m_arrayElements * 3];
    calculateMomentumViewPositions(m_momentumViewPositions);

    // Initialise Borders
    m_staticBorderPotential = new float[m_arrayElements];
    calculateBorderPotential(m_staticBorderPotential);
    m_dynamicPotential = new float[m_arrayElements];
    glGenBuffers(1, &m_potentialBuffer);
    glBindBuffer(GL_ARRAY_BUFFER, m_potentialBuffer);
    glBufferData(GL_ARRAY_BUFFER, sizeof(float) * m_arrayElements, m_staticBorderPotential, GL_STATIC_DRAW);
    glEnableVertexAttribArray(2);
    glVertexAttribPointer(2, 1, GL_FLOAT, GL_FALSE, sizeof(float), 0);

    // Set orthografic view projectin "camera" which will not change.
    m_viewProjectionPosition = glm::ortho(-3.2f, 3.2f, -1.8f, 1.8f, -1.0f, 1.0f);
    m_viewProjectionLocation = glGetUniformLocation(m_shader->getRendererID(), "u_viewProjection");
    m_showPotentialsLocation = glGetUniformLocation(m_shader->getRendererID(), "u_showPotential");

    fftw_init_threads();
    m_planForward = fftw_plan_dft_2d(m_arraySize, m_arraySize, (fftw_complex *)m_psi, (fftw_complex *)m_psi, FFTW_FORWARD, FFTW_MEASURE);
    m_planBackwards = fftw_plan_dft_2d(m_arraySize, m_arraySize, (fftw_complex *)m_psi, (fftw_complex *)m_psi, FFTW_BACKWARD, FFTW_MEASURE);
    
    initialiseParticleMomentum();
}

void QPong::GameLayer::onDetach()
{   
    glDeleteVertexArrays(1, &m_vertexArray);
    glDeleteBuffers(1, &m_pointsBuffer);
    glDeleteBuffers(1, &m_indicesBuffer);
    glDeleteBuffers(1, &m_colorBuffer);
    fftw_free(m_psi);
    delete(m_momentum);
    delete(m_dynamicPotential);
    delete(m_staticBorderPotential);
}

void QPong::GameLayer::onEvent(GuiCore::Event& event)
{
}

void QPong::GameLayer::reset()
{
    initialiseParticleMomentum();
    m_restart = false;
    m_gameCounts = true;
    m_playerOne.currentlyScored = 0.0;
    m_playerTwo.currentlyScored = 0.0;

}

void QPong::GameLayer::onUpdate(GuiCore::Timestep ts)
{
    GuiCore::Glyphs::renderText("QPong", -0.31f, 1.59f, 0.005f, glm::vec3(0.5f, 0.9f, 0.9f));
    if (m_restart)
    {
        reset();
    }
    if (m_propagate)
    {
        m_leftBat.onUpdate(ts);
        m_rightBat.onUpdate(ts);
        propagateStep(ts.getSeconds() * m_speedScale);

        if (m_gameCounts)
        {
            if (m_playerOne.currentlyScored > 35.0)
            {
                m_playerOne.totalScore++;
                m_gameCounts = false;
                m_propagate = false;
            }
            if (m_playerTwo.currentlyScored > 35.0)
            {
                m_playerTwo.totalScore++;
                m_gameCounts = false;
                m_propagate = false;
            }
        }
    }
    
    // Particle
    glBindBuffer(GL_ARRAY_BUFFER, m_pointsBuffer);
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(float) * m_arrayElements * 3, m_particleViewPositions);
    
    glBindBuffer(GL_ARRAY_BUFFER, m_colorBuffer);
    glBufferSubData(GL_ARRAY_BUFFER, 0, m_memorySize, m_psi);
    
    glBindBuffer(GL_ARRAY_BUFFER, m_potentialBuffer);
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(float) * m_arrayElements, m_dynamicPotential);

    glUseProgram(m_shader->getRendererID());
    glUniform1i(m_showPotentialsLocation, 1);
    glUniformMatrix4fv(m_viewProjectionLocation, 1, GL_FALSE, &m_viewProjectionPosition[0][0]);
    glBindVertexArray(m_vertexArray);
    glDrawElements(GL_TRIANGLES, m_numberOfIndices, GL_UNSIGNED_INT, nullptr);

    // Momentum
    if (m_renderMomentum)
    {    
        glBindBuffer(GL_ARRAY_BUFFER, m_pointsBuffer);
        glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(float) * m_arrayElements * 3, m_momentumViewPositions);
        
        glBindBuffer(GL_ARRAY_BUFFER, m_colorBuffer);
        glBufferSubData(GL_ARRAY_BUFFER, 0, m_memorySize, m_momentum);
        
        glUseProgram(m_shader->getRendererID());
        glUniform1i(m_showPotentialsLocation, 0);
        glBindVertexArray(m_vertexArray);
        glDrawElements(GL_TRIANGLES, m_numberOfIndices, GL_UNSIGNED_INT, nullptr);
    }

    std::stringstream totalScore;
    totalScore << m_playerOne.totalScore << " : " << m_playerTwo.totalScore;
    GuiCore::Glyphs::renderText(totalScore.str(), -1.82f, 1.52f, 0.0075f, {1.0, 1.0, 1.0});
}

void QPong::GameLayer::onGuiRender()
{
    // GUI settings for the particle
    ImGui::Begin("Particle Settings");
    
    // Select initial momentum for particle at restart
    ImGui::DragFloat2("Momentum [x,y]", glm::value_ptr(m_initialParticleProperties.momentum), 0.01f, -3.0f, 3.0f);
    
    float pointUnsharpness = m_initialParticleProperties.pointUnsharpness;
    if (ImGui::DragFloat("Sharpness", &pointUnsharpness, 1.0f, 5.0f, 1000.0f))
    {
        m_initialParticleProperties.pointUnsharpness = pointUnsharpness;
    }
    // Select hbar, changes apply on the fly.
    float hbar = m_initialParticleProperties.hbar;
    if (ImGui::DragFloat("hbar", &hbar, 0.000001f, 0.000001f, 1.0f, "%.6f"))
    {
        m_initialParticleProperties.hbar = hbar;
    }

    // This value increases or decreases the time step made every iteration.
    ImGui::DragFloat("Speed Factor", &m_speedScale, 0.1, 0.1, 20.0);
    if (ImGui::IsItemHovered()) {
        ImGui::SetTooltip("Change the size of the timestep (dt)\nfor every iteration of the simulation.");
    }

    // Decrease the 
    if (ImGui::Button("Less threads"))
    {
        if (m_renderingThreadsNumber > 1)
        {
            bool isPowerOf2 = false;
            while (!isPowerOf2)
            {
                m_renderingThreadsNumber--;
                for (int i = 0; i < 10; i++)
                {
                    if (m_renderingThreadsNumber == pow(2, i))
                    {
                        isPowerOf2 = true;
                        break;
                    }
                }
            }
        }
    }
    ImGui::SameLine();
    if (ImGui::Button("More Threads"))
    {
        if (m_renderingThreadsNumber <  200)
        {
            bool isPowerOf2 = false;
            while (!isPowerOf2)
            {
                m_renderingThreadsNumber++;
                for (int i = 0; i < 10; i++)
                {
                    if (m_renderingThreadsNumber == pow(2, i))
                    {
                        isPowerOf2 = true;
                        break;
                    }
                }
            }
        }
    }
    ImGui::SameLine();
    ImGui::Text("%d", m_renderingThreadsNumber);
    ImGui::Checkbox("Render Momentum", &m_renderMomentum);
    ImGui::Checkbox("Enable Potentials", &m_potentialsEnabled);
    ImGui::Checkbox("Enable Particle absorption", &m_removeParticleEnabled);
    // Start/Stop propagating
    if (m_propagate)
    {
        if (ImGui::Button("Pause"))
        {
            m_propagate = false;
        }
    }
    else
    {
        if (ImGui::Button("Play"))
        {
            m_propagate = true;
        }
    }
    // Restart game
    if (ImGui::Button("Restart"))
    {
        m_restart = true;
    }
    ImGui::Text("Player 1: %d", m_playerOne.totalScore); ImGui::SameLine();
    ImGui::Text(" | Live Score: %.1f\%%",  m_playerOne.currentlyScored);
    
    ImGui::Text("Player 2: %d", m_playerTwo.totalScore); ImGui::SameLine();
    ImGui::Text(" | Live Score: %.1f\%%",  m_playerTwo.currentlyScored);

    ImGui::End();

}