/// @file   GameLayer.hpp
/// @author Armin Co
///

#ifndef QPONG_GAME_LAYER_HPP
#define QPONG_GAME_LAYER_HPP

#include <glm/glm.hpp>

#include <GuiCore/Layer.hpp>
#include <GuiCore/Shader.hpp>
#include <GuiCore/Timestep.hpp>

#include <complex>
#include <fftw3.h>

namespace QPong
{
class GameLayer : public GuiCore::Layer
{
public:
    struct Properties
    {
        int resolution;
        Properties(int res) : resolution {res}{};
    };

    struct ParticleProps
    {
        glm::vec3 position;
        glm::vec3 momentum;
        double pointUnsharpness;
        double startValue;
        double hbar;
    };

    struct Player
    {
        double currentlyScored;
        int totalScore;
    };

    class Bat
    {
    public:
        Bat(double x, double y, int keyUp, int keyDown);
        double getX() const;
        double getY() const;
        void onUpdate(GuiCore::Timestep ts);
        double getPotential(double x, double y) const;
    private:
        double m_x;
        double m_y;
        double m_vx;
        double m_vy;
        double m_height;
        double m_width;
        int m_keyUp;
        int m_keyDown;
        static double s_speedStepSize;
    };

    GameLayer(Properties props);
    virtual ~GameLayer(){};
    virtual void onAttach() override;
    virtual void onDetach() override;
    virtual void onUpdate(GuiCore::Timestep ts) override;
    virtual void onGuiRender() override;
    virtual void onEvent(GuiCore::Event &event) override;

private:
    void reset();

    void calculatePointPositions(float *points);
    void calculateMomentumViewPositions(float *points);
    void calculateTriangleIndices(uint32_t *indices);
    void calculateBorderPotential(float *potential);
    void initialiseParticleMomentum();
    void propagateStep(double dt);
    void applyPotentials(int indexBegin, int indexEnd, double dt);
    void moveForward(int indexBegin, int indexEnd, double dt);
    void updateMomentumView(int indexBegin, int indexEnd);
    double absorbParticle(int i, double c);
    // positions at array index
    inline double posAt(int pos) const;
    inline double xAtIndex(int i) const;
    inline double yAtIndex(int i) const;
    inline double potentialAt(int pos) const;
    inline double pxAtIndex(int i) const;
    inline double pyAtIndex(int i) const;

    GuiCore::Shader *m_shader;
    std::string m_vertexShaderPath;
    std::string m_fragmentShaderPath;

    int m_arraySize;
    int m_arrayElements;
    int m_numberOfQuads;
    int m_numberOfIndices;

    GLuint m_vertexArray;       ///< OpenGL object that "knows" all buffers.
    GLuint m_pointsBuffer;      ///< Buffer with the location for each element in the particle array
    GLuint m_indicesBuffer;     ///< Indicies to draw trinagles from the points to fill the "canvas"
    GLuint m_colorBuffer;       ///< Reference for OpenGL to load the paritcle data to the GPU
    GLuint m_potentialBuffer;   ///< Borders and bats

    float *m_particleViewPositions;
    float *m_momentumViewPositions;
    float *m_staticBorderPotential;
    float *m_dynamicPotential;

    glm::mat4 m_viewProjectionPosition;
    int m_viewProjectionLocation;
    GLuint m_showPotentialsLocation;

    ParticleProps m_initialParticleProperties;
    double m_particleAbsouluteSumStart;
    fftw_plan m_planForward;
    fftw_plan m_planBackwards;

    Bat m_leftBat;
    Bat m_rightBat;
    Player m_playerOne;
    Player m_playerTwo;
    bool m_gameCounts;

    bool m_restart;
    bool m_propagate;
    bool m_renderMomentum;
    bool m_removeParticleEnabled;
    bool m_potentialsEnabled;
    int m_renderingThreadsNumber {8};
    size_t m_memorySize {0};
    float m_speedScale {1.0f};
    std::complex<double> *m_psi;
    std::complex<double> *m_momentum;
};
}
#endif