/// @file   Window.cpp
/// @author Armin Co
///


#include <GL/gl3w.h> 
#include <GuiCore/Log.hpp>
#include <GuiCore/ApplicationEvent.hpp>

#include "Window.hpp"


static bool s_GLFW_Initialized = false;

static void GLFWErrorCallback(int error, const char *description)
{
    /// @todo add GuiCore::Logger
}

GuiCore::IWindow* GuiCore::IWindow::create(const WindowProperties &properties)
{
    return new Window(properties);
}

Window::Window(const GuiCore::WindowProperties &properties)
{
    setup(properties);
}

Window::~Window()
{
    shutdown();
}

void Window::setup(const GuiCore::WindowProperties &properties)
{
    m_data.title = properties.title;
    m_data.width = properties.width;
    m_data.height = properties.height;

    if (!s_GLFW_Initialized)
    {
        glfwSetErrorCallback(GLFWErrorCallback);
        auto success = glfwInit();
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);
        glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE); 

        s_GLFW_Initialized = true;
    }

    m_window = glfwCreateWindow(properties.width, properties.height, m_data.title.c_str(), nullptr, nullptr);
    
    glfwMakeContextCurrent(m_window);
    glfwSetWindowUserPointer(m_window, &m_data);

    setVSync(true);

    gl3wInit();

    glfwSetWindowSizeCallback(m_window, [](GLFWwindow * window, int width, int height)
    {
        Data &data = *(Data*)glfwGetWindowUserPointer(window);
        data.width = width;
        data.height = height;

        GuiCore::WindowResizeEvent event(width, height);
        data.eventCallback(event);
    });

    using namespace GuiCore;
    glfwSetWindowCloseCallback(m_window, [](GLFWwindow* window)
    {
        Data &data = *(Data*)glfwGetWindowUserPointer(window);
        WindowCloseEvent event;
        data.eventCallback(event);
    });
}

void Window::shutdown()
{
    glfwDestroyWindow(m_window);
}

void Window::onUpdate()
{
    glfwPollEvents();
    glfwSwapBuffers(m_window);
}

void Window::setVSync(bool enabled)
{
    if (enabled)
    {
        glfwSwapInterval(1);
    }
    else
    {
        glfwSwapInterval(0);
    }
    m_data.vSyncEnabled = enabled;
}

bool Window::isVSyncEnabled() const
{
    return m_data.vSyncEnabled;
}

uint32_t Window::getWidth() const
{
    return m_data.width;
}

uint32_t Window::getHeight() const
{
    return m_data.height;
}

void* Window::getNativeWindow() const
{
    return m_window;
}

void Window::setEventCallback(const std::function<void(GuiCore::Event&)> callback)
{
    GuiCore::Log::get()->trace("Window::setEventCallback");
    m_data.eventCallback = callback;
}